'''This module contains all high-level helpers function that allow to work with
Cyclomatic Complexity
'''

import math

from radon.visitors import GET_COMPLEXITY, ComplexityVisitor, code2ast

# sorted_block ordering functions
SCORE = lambda block: -GET_COMPLEXITY(block)
LINES = lambda block: block.lineno
ALPHA = lambda block: block.name


def cc_rank(cc):
    r'''Rank the complexity score from A to F, where A stands for the simplest
    and best score and F the most complex and worst one:

    ============= =====================================================
        1 - 5        A (low risk - simple block)
        6 - 10       B (low risk - well structured and stable block)
        11 - 20      C (moderate risk - slightly complex block)
        21 - 30      D (more than moderate risk - more complex block)
        31 - 40      E (high risk - complex block, alarming)
        41+          F (very high risk - error-prone, unstable block)
    ============= =====================================================

    Here *block* is used in place of function, method or class.

    The formula used to convert the score into an index is the following:

    .. math::

        \text{rank} = \left \lceil \dfrac{\text{score}}{10} \right \rceil
        - H(5 - \text{score})

    where ``H(s)`` stands for the Heaviside Step Function.
    The rank is then associated to a letter (0 = A, 5 = F).
    '''
    if cc < 0:
        raise ValueError('Complexity must be a non-negative value')
    return chr(
        min(int(math.ceil(cc / 10.0) or 1) - (1, 0)[5 - cc < 0], 5) + 65
    )


def average_complexity(blocks):
    '''Compute the average Cyclomatic complexity from the given blocks.
    Blocks must be either :class:`~radon.visitors.Function` or
    :class:`~radon.visitors.Class`. If the block list is empty, then 0 is
    returned.
    '''
    size = len(blocks)
    if size == 0:
        return 0
    return sum((GET_COMPLEXITY(block) for block in blocks), 0.0) / len(blocks)


def sorted_results(blocks, order=SCORE):
    '''Given a ComplexityVisitor instance, returns a list of sorted blocks
    with respect to complexity. A block is a either
    :class:`~radon.visitors.Function` object or a
    :class:`~radon.visitors.Class` object.
    The blocks are sorted in descending order from the block with the highest
    complexity.

    The optional `order` parameter indicates how to sort the blocks. It can be:

        * `LINES`: sort by line numbering;
        * `ALPHA`: sort by name (from A to Z);
        * `SCORE`: sorty by score (descending).

    Default is `SCORE`.
    '''
    return sorted(blocks, key=order)


def add_inner_blocks(blocks):
    '''Process a list of blocks by adding all closures and inner classes as
    top-level blocks.
    '''
    new_blocks = []
    all_blocks = blocks[:]
    while all_blocks:
        block = all_blocks.pop()
        new_blocks.append(block)
        for inner_block in ('closures', 'inner_classes'):
            for i_block in getattr(block, inner_block, ()):
                named = i_block._replace(name=block.name + '.' + i_block.name)
                all_blocks.append(named)
                for meth in getattr(named, 'methods', ()):
                    m_named = meth._replace(classname=named.name)
                    all_blocks.append(m_named)
    return new_blocks


def cc_visit(code, **kwargs):
    '''Visit the given code with :class:`~radon.visitors.ComplexityVisitor`.
    All the keyword arguments are directly passed to the visitor.
    '''
    return cc_visit_ast(code2ast(code), **kwargs)


def cc_visit_ast(ast_node, **kwargs):
    '''Visit the AST node with :class:`~radon.visitors.ComplexityVisitor`. All
    the keyword arguments are directly passed to the visitor.
    '''
    return ComplexityVisitor.from_ast(ast_node, **kwargs).blocks



import json
import os

import pytest

import radon.cli as cli
from radon.cli.harvest import (
    SUPPORTS_IPYNB,
    CCHarvester,
    Harvester,
    MIHarvester,
    RawHarvester,
)
from radon.cli.tools import _is_python_file
from radon.tests.test_cli_harvest import MI_CONFIG, RAW_CONFIG

BASE_CONFIG_WITH_IPYNB = cli.Config(
    exclude="*.py", ignore=None, include_ipynb=True, ipynb_cells=False,
)

BASE_CONFIG_WITH_IPYNB_AND_CELLS = cli.Config(
    exclude="*.py", ignore=None, include_ipynb=True, ipynb_cells=True,
)

DIRNAME = os.path.dirname(__file__)


@pytest.mark.skipif(not SUPPORTS_IPYNB, reason="nbformat not installed")
def test_harvestor_yields_ipynb(log_mock):
    '''Test that Harvester will try ipynb files when configured'''
    target = os.path.join(DIRNAME, 'data/example.ipynb')
    harvester = Harvester([DIRNAME], BASE_CONFIG_WITH_IPYNB)
    filenames = list(harvester._iter_filenames())
    assert _is_python_file(target)
    assert len(filenames) == 1
    assert target in filenames


@pytest.mark.skipif(not SUPPORTS_IPYNB, reason="nbformat not installed")
def test_ipynb(log_mock):
    mi_cfg = cli.Config(**BASE_CONFIG_WITH_IPYNB.config_values)
    mi_cfg.config_values.update(MI_CONFIG.config_values)
    raw_cfg = cli.Config(**BASE_CONFIG_WITH_IPYNB.config_values)
    raw_cfg.config_values.update(RAW_CONFIG.config_values)
    cc_cfg = cli.Config(
        order=lambda block: block.name,
        no_assert=False,
        min='A',
        max='F',
        show_complexity=False,
        show_closures=False,
        average=True,
        total_average=False,
    )
    cc_cfg.config_values.update(BASE_CONFIG_WITH_IPYNB.config_values)

    mappings = {
        MIHarvester: mi_cfg,
        RawHarvester: raw_cfg,
        CCHarvester: cc_cfg,
    }
    target = 'data/'
    fnames = [
        os.path.join(DIRNAME, target),
    ]
    for h_class, cfg in mappings.items():
        for f in fnames:
            harvester = h_class([f], cfg)
            out = harvester.as_json()
            assert not any(['error' in out]), out


@pytest.mark.skipif(not SUPPORTS_IPYNB, reason="nbformat not installed")
def test_ipynb_with_cells(mocker, log_mock):
    mi_cfg = cli.Config(**BASE_CONFIG_WITH_IPYNB_AND_CELLS.config_values)
    mi_cfg.config_values.update(MI_CONFIG.config_values)
    raw_cfg = cli.Config(**BASE_CONFIG_WITH_IPYNB_AND_CELLS.config_values)
    raw_cfg.config_values.update(RAW_CONFIG.config_values)
    cc_cfg = cli.Config(
        order=lambda block: block.name,
        no_assert=False,
        min='A',
        max='F',
        show_complexity=False,
        show_closures=False,
        average=True,
        total_average=False,
    )
    cc_cfg.config_values.update(BASE_CONFIG_WITH_IPYNB_AND_CELLS.config_values)

    mappings = {
        MIHarvester: mi_cfg,
        RawHarvester: raw_cfg,
        CCHarvester: cc_cfg,
    }
    target = 'data/'
    fnames = [
        os.path.join(DIRNAME, target),
    ]
    for h_class, cfg in mappings.items():
        for f in fnames:
            harvester = h_class([f], cfg)
            out = harvester.as_json()
            assert not any(['error' in out]), out


@pytest.mark.skipif(not SUPPORTS_IPYNB, reason="nbformat not installed")
def test_raw_ipynb(log_mock):
    raw_cfg = cli.Config(**BASE_CONFIG_WITH_IPYNB.config_values)

    target = os.path.join(DIRNAME, 'data/example.ipynb')
    harvester = RawHarvester([DIRNAME], raw_cfg)
    out = json.loads(harvester.as_json())
    assert harvester.config.include_ipynb == True
    assert target in out
    assert out[target]['loc'] == 63
    assert out[target]['lloc'] == 37
    assert out[target]['sloc'] == 37
    assert out[target]['comments'] == 3
    assert out[target]['multi'] == 10
    assert out[target]['blank'] == 14
    assert out[target]['single_comments'] == 2


@pytest.mark.skipif(not SUPPORTS_IPYNB, reason="nbformat not installed")
def test_raw_ipynb_cells(log_mock):
    raw_cfg = cli.Config(**BASE_CONFIG_WITH_IPYNB_AND_CELLS.config_values)

    target = os.path.join(DIRNAME, 'data/example.ipynb')
    harvester = RawHarvester([DIRNAME], raw_cfg)
    out = json.loads(harvester.as_json())
    cell_target = target + ":[3]"
    assert target in out
    assert cell_target in out
    assert out[cell_target]['loc'] == 52
    assert out[cell_target]['lloc'] == 27
    assert out[cell_target]['sloc'] == 27
    assert out[cell_target]['comments'] == 3
    assert out[cell_target]['multi'] == 10
    assert out[cell_target]['blank'] == 13
    assert out[cell_target]['single_comments'] == 2



'''Module holding constants used to format lines that are printed to the
terminal.
'''

import sys

try:
    import colorama

    colorama.init(strip=(not sys.stdout.isatty()))
    GREEN, YELLOW, RED = (
        colorama.Fore.GREEN,
        colorama.Fore.YELLOW,
        colorama.Fore.RED,
    )
    MAGENTA, CYAN, WHITE = (
        colorama.Fore.MAGENTA,
        colorama.Fore.CYAN,
        colorama.Fore.WHITE,
    )
    BRIGHT, RESET = colorama.Style.BRIGHT, colorama.Style.RESET_ALL
except ImportError:  # pragma: no cover
    # No colorama, so let's fallback to no-color mode
    GREEN = YELLOW = RED = MAGENTA = CYAN = WHITE = BRIGHT = RESET = ''

RANKS_COLORS = {
    'A': GREEN,
    'B': GREEN,
    'C': YELLOW,
    'D': YELLOW,
    'E': RED,
    'F': RED,
}

LETTERS_COLORS = {'F': MAGENTA, 'C': CYAN, 'M': WHITE}

MI_RANKS = {'A': GREEN, 'B': YELLOW, 'C': RED}
TEMPLATE = '{0}{1} {reset}{2}:{3} {4} - {5}{6}{reset}'



import sys
import textwrap

import pytest

from radon.visitors import HalsteadVisitor

dedent = lambda code: textwrap.dedent(code).strip()


SIMPLE_BLOCKS = [
    (
        '''
     if a and b: pass
     ''',
        (1, 2, 1, 2),
    ),
    (
        '''
     if a and b: pass
     elif b or c: pass
     ''',
        (2, 4, 2, 3),
    ),
    (
        '''
     if a and b: pass
     elif b and c: pass
     ''',
        (2, 4, 1, 3),
    ),
    (
        '''
     a = b * c
     ''',
        (1, 2, 1, 2),
    ),
    (
        '''
     b = -x
     ''',
        (1, 1, 1, 1),
    ),
    (
        '''
     a = -x
     c = -x
     ''',
        (2, 2, 1, 1),
    ),
    (
        '''
     a = -x
     b = +x
     ''',
        (2, 2, 2, 1),
    ),
    (
        '''
     a += 3
     b += 4
     c *= 3
     ''',
        (3, 6, 2, 5),
    ),
    (
        '''
     a = 2
     b = 3
     a *= b

     def f():
         b = 2
         b += 4
     ''',
        (2, 4, 2, 4),
    ),
    (
        '''
     a = b < 4
     c = i <= 45 >= d
     k = 4 < 2
     ''',
        (4, 7, 3, 6),
    ),
]

if sys.version_info[:2] >= (3, 5):
    SIMPLE_BLOCKS.append(
        (
            '''
        a = 2
        b = 3
        a *= b

        async def f():
            b = 2
            b += 4
        ''',
            (2, 4, 2, 4),
        ),
    )


@pytest.mark.parametrize('code,expected', SIMPLE_BLOCKS)
def test_visitor(code, expected):
    visitor = HalsteadVisitor.from_code(dedent(code))
    assert expected == (
        visitor.operators,
        visitor.operands,
        visitor.distinct_operators,
        visitor.distinct_operands,
    )



from flake8_polyfill import options

from radon.complexity import add_inner_blocks
from radon.visitors import ComplexityVisitor


class Flake8Checker(object):
    '''Entry point for the Flake8 tool.'''

    name = 'radon'
    version = __import__('radon').__version__
    _code = 'R701'
    _error_tmpl = 'R701 %r is too complex (%d)'
    no_assert = False
    show_closures = False
    max_cc = -1

    def __init__(self, tree, filename):
        '''Accept the AST tree and a filename (unused).'''
        self.tree = tree

    @classmethod
    def add_options(cls, parser):  # pragma: no cover
        '''Add custom options to the global parser.'''
        options.register(
            parser,
            '--radon-max-cc',
            default=-1,
            action='store',
            type='int',
            help='Radon complexity threshold',
            parse_from_config=True,
        )
        options.register(
            parser,
            '--radon-no-assert',
            dest='no_assert',
            action='store_true',
            default=False,
            help='Radon will ignore assert statements',
            parse_from_config=True,
        )
        options.register(
            parser,
            '--radon-show-closures',
            dest='show_closures',
            action='store_true',
            default=False,
            help='Add closures/inner classes to the output',
            parse_from_config=True,
        )

    @classmethod
    def parse_options(cls, options):  # pragma: no cover
        '''Save actual options as class attributes.'''
        cls.max_cc = options.radon_max_cc
        cls.no_assert = options.no_assert
        cls.show_closures = options.show_closures

    def run(self):
        '''Run the ComplexityVisitor over the AST tree.'''
        if self.max_cc < 0:
            if not self.no_assert:
                return
            self.max_cc = 10
        visitor = ComplexityVisitor.from_ast(
            self.tree, no_assert=self.no_assert
        )

        blocks = visitor.blocks
        if self.show_closures:
            blocks = add_inner_blocks(blocks)

        for block in blocks:
            if block.complexity > self.max_cc:
                text = self._error_tmpl % (block.name, block.complexity)
                yield block.lineno, block.col_offset, text, type(self)



import os
import sys
from configparser import ConfigParser

import pytest

import radon.cli as cli
import radon.complexity as cc_mod
from radon.cli.harvest import CCHarvester, Harvester, MIHarvester, RawHarvester
from radon.tests.test_cli_harvest import (
    BASE_CONFIG,
    CC_CONFIG,
    MI_CONFIG,
    RAW_CONFIG,
)

DIRNAME = os.path.dirname(__file__)


def func(a, b=2, c=[], d=None):
    pass


def func2(*args, **kwargs):
    pass


def func3(b=3, *args):
    pass


def fake_to_terminal():
    yield ('a', ('mystr',), {'error': True})
    yield ('b', (), {})
    yield (('p1', 'p2'), (), {'indent': 1})


def test_config_base_behavior():
    c = cli.Config(a=2, b=3)
    assert c.config_values == {'a': 2, 'b': 3}
    assert c.a == 2
    assert c.b == 3


def test_config_exceptions():
    c = cli.Config(a=2)
    assert c.__dict__, {'config_values': {'a': 2}}
    with pytest.raises(AttributeError):
        c.notexistent


def test_config_str():
    assert str(cli.Config()) == '{}'
    assert str(cli.Config(a=2)) == '{\'a\': 2}'


def test_config_eq():
    assert cli.Config() == cli.Config()
    assert cli.Config(a=2) == cli.Config(a=2)
    assert cli.Config(a=2) != cli.Config(b=2)


def test_config_for():
    assert cli.Config.from_function(func) == cli.Config(b=2, c=[], d=None)
    assert cli.Config.from_function(func2) == cli.Config()
    assert cli.Config.from_function(func3) == cli.Config(b=3)


def test_config_converts_types(mocker):
    test_config = ConfigParser()
    test_config.read_string(
        u'''
        [radon]
        str_test = B
        int_test = 19
        bool_test = true
        '''
    )
    config_mock = mocker.patch('radon.cli.FileConfig.file_config')
    config_mock.return_value = test_config

    cfg = cli.FileConfig()
    assert cfg.get_value('bool_test', bool, False) == True
    assert cfg.get_value('str_test', str, 'x') == 'B'
    assert cfg.get_value('missing_test', str, 'Y') == 'Y'
    assert cfg.get_value('int_test', int, 10) == 19


def test_cc(mocker, log_mock):
    harv_mock = mocker.patch('radon.cli.CCHarvester')
    harv_mock.return_value = mocker.sentinel.harvester

    cli.cc(['-'], json=True)

    harv_mock.assert_called_once_with(
        ['-'],
        cli.Config(
            min='A',
            max='F',
            exclude=None,
            ignore=None,
            show_complexity=False,
            average=False,
            order=getattr(cc_mod, 'SCORE'),
            no_assert=False,
            total_average=False,
            show_closures=False,
            include_ipynb=False,
            ipynb_cells=False,
        ),
    )
    log_mock.assert_called_once_with(
        mocker.sentinel.harvester,
        codeclimate=False,
        json=True,
        stream=sys.stdout,
        xml=False,
        md=False
    )


def test_raw(mocker, log_mock):
    harv_mock = mocker.patch('radon.cli.RawHarvester')
    harv_mock.return_value = mocker.sentinel.harvester

    cli.raw(['-'], summary=True, json=True)

    harv_mock.assert_called_once_with(
        ['-'],
        cli.Config(
            exclude=None,
            ignore=None,
            summary=True,
            include_ipynb=False,
            ipynb_cells=False,
        ),
    )
    log_mock.assert_called_once_with(
        mocker.sentinel.harvester, stream=sys.stdout, json=True
    )


def test_mi(mocker, log_mock):
    harv_mock = mocker.patch('radon.cli.MIHarvester')
    harv_mock.return_value = mocker.sentinel.harvester

    cli.mi(['-'], show=True, multi=False)

    harv_mock.assert_called_once_with(
        ['-'],
        cli.Config(
            min='A',
            max='C',
            exclude=None,
            ignore=None,
            show=True,
            multi=False,
            sort=False,
            include_ipynb=False,
            ipynb_cells=False,
        ),
    )
    log_mock.assert_called_once_with(
        mocker.sentinel.harvester, stream=sys.stdout, json=False
    )


def test_encoding(mocker, log_mock):
    mi_cfg = cli.Config(**BASE_CONFIG.config_values)
    mi_cfg.config_values.update(MI_CONFIG.config_values)
    raw_cfg = cli.Config(**BASE_CONFIG.config_values)
    raw_cfg.config_values.update(RAW_CONFIG.config_values)
    mappings = {
        MIHarvester: mi_cfg,
        RawHarvester: raw_cfg,
        CCHarvester: CC_CONFIG,
    }
    if sys.version_info[0] < 3:
        target = 'data/__init__.py'
    else:
        target = 'data/py3unicode.py'
    fnames = [
        os.path.join(DIRNAME, target),
        # This one will fail if detect_encoding() removes the first lines
        # See #133
        os.path.join(DIRNAME, 'data/no_encoding.py'),
    ]
    for h_class, cfg in mappings.items():
        for f in fnames:
            harvester = h_class([f], cfg)
            assert not any(
                ['error' in kw for msg, args, kw in harvester.to_terminal()]
            )


@pytest.fixture
def stdout_mock(mocker):
    return mocker.patch('radon.cli.sys.stdout.write')


def test_log(mocker, stdout_mock):
    cli.log('msg')
    cli.log('msg', indent=1)
    cli.log('{0} + 1', 2)
    cli.log('{0} + 1', 2, noformat=True)

    stdout_mock.assert_has_calls(
        [
            mocker.call('msg\n'),
            mocker.call('    msg\n'),
            mocker.call('2 + 1\n'),
            mocker.call('{0} + 1\n'),
        ]
    )
    assert stdout_mock.call_count == 4


def test_log_list(stdout_mock):
    cli.log_list([])
    cli.log_list(['msg'])

    stdout_mock.assert_called_once_with('msg\n')


def test_log_error(mocker, stdout_mock):
    reset_mock = mocker.patch('radon.cli.RESET')
    red_mock = mocker.patch('radon.cli.RED')
    bright_mock = mocker.patch('radon.cli.BRIGHT')

    bright_mock.__str__.return_value = '@'
    red_mock.__str__.return_value = '<|||>'
    reset_mock.__str__.return_value = '__R__'

    cli.log_error('mystr')

    stdout_mock.assert_called_once_with('@<|||>ERROR__R__: mystr\n')


def test_log_result(mocker, stdout_mock):
    le_mock = mocker.patch('radon.cli.log_error')
    ll_mock = mocker.patch('radon.cli.log_list')
    log_mock = mocker.patch('radon.cli.log')

    h = mocker.Mock(spec=Harvester)
    h.as_json.return_value = mocker.sentinel.json
    h.as_xml.return_value = mocker.sentinel.xml
    h.as_md.return_value = mocker.sentinel.md
    h.to_terminal.side_effect = fake_to_terminal

    cli.log_result(h, json=True)
    h.as_json.assert_called_once_with()

    h.as_json.reset_mock()
    cli.log_result(h, json=True, xml=True, md=True)
    h.as_json.assert_called_once_with()
    assert h.as_xml.call_count == 0
    assert h.as_md.call_count == 0

    cli.log_result(h, xml=True)
    h.as_xml.assert_called_once_with()

    cli.log_result(h, md=True)
    h.as_md.assert_called_once_with()

    cli.log_result(h)
    h.to_terminal.assert_called_once_with()

    log_mock.assert_has_calls(
        [
            mocker.call(mocker.sentinel.json, json=True, noformat=True),
            mocker.call(
                mocker.sentinel.json, json=True, noformat=True, xml=True, md=True
            ),
            mocker.call(mocker.sentinel.xml, noformat=True, xml=True),
            mocker.call(mocker.sentinel.md, noformat=True, md=True),
            mocker.call('a', error=True),
        ]
    )
    le_mock.assert_called_once_with('mystr', indent=1)
    ll_mock.assert_has_calls(
        [mocker.call(['b']), mocker.call(('p1', 'p2'), indent=1)]
    )



import json
import locale
import os
import platform
import sys

import pytest

import radon.cli.tools as tools
from radon.raw import Module
from radon.visitors import Class, Function


def fake_isfile(filename):
    if filename == 'file.py':
        return True
    return False


def fake_walk(start):
    dirs = ['tests', 'sub', '.hid']
    contents = {
        'tests': ['test_amod.py', 'run.py', '.hid.py'],
        'sub': ['amod.py', 'bmod.py'],
    }
    yield '.', dirs, ['tox.ini', 'amod.py', 'test_all.py', 'fake.yp', 'noext']
    for d in dirs:
        yield './{0}'.format(d), [], contents[d]


def fake_is_python_file(filename):
    return filename.endswith('.py')


def assert_pequal(a, b):
    a, b = [list(map(os.path.normpath, p)) for p in (a, b)]
    assert a == b


def test_open(mocker):
    with tools._open('-') as fobj:
        assert fobj is sys.stdin

    try:
        with tools._open(__file__) as fobj:
            assert True
    except TypeError:  # issue 101
        assert False, 'tools._open raised TypeError'

    m = mocker.mock_open()

    if platform.python_implementation() == 'PyPy':
        mocker.patch('radon.cli.tools.open', m, create=True)
        tools._open('randomfile.py').__enter__()
        m.assert_called_with('randomfile.py')
    else:
        mocker.patch('radon.cli.tools._open_function', m, create=True)
        tools._open('randomfile.py').__enter__()
        if sys.version_info[:2] >= (3, 0):
            default_encoding = 'utf-8'
        else:
            default_encoding = locale.getpreferredencoding(False)
        except_encoding = os.getenv(
            'RADONFILESENCODING', default_encoding
        )
        m.assert_called_with('randomfile.py', encoding=except_encoding)


@pytest.fixture
def iter_files():
    return lambda *a, **kw: list(tools.iter_filenames(*a, **kw))


def test_iter_files_stdin(iter_files):
    assert iter_files(['-']) == ['-']


def test_iter_files(mocker, iter_files):
    os_mod = mocker.patch('radon.cli.tools.os')
    os_path_mod = mocker.patch('radon.cli.tools.os.path')
    os_path_mod.normpath = os.path.normpath
    os_path_mod.basename = os.path.basename
    os_path_mod.join = os.path.join

    os_path_mod.isfile.side_effect = fake_isfile
    _orig_walk = os_mod.walk
    os_mod.walk = fake_walk
    _orig_is_python_file = tools._is_python_file
    tools._is_python_file = fake_is_python_file

    assert_pequal(
        iter_files(['file.py', 'random/path']),
        [
            'file.py',
            'amod.py',
            'test_all.py',
            'tests/test_amod.py',
            'tests/run.py',
            'sub/amod.py',
            'sub/bmod.py',
        ],
    )

    assert_pequal(
        iter_files(['file.py', 'random/path'], 'test_*'),
        [
            'file.py',
            'amod.py',
            'tests/test_amod.py',
            'tests/run.py',
            'sub/amod.py',
            'sub/bmod.py',
        ],
    )

    assert_pequal(
        iter_files(['file.py', 'random/path'], '*test_*'),
        ['file.py', 'amod.py', 'tests/run.py', 'sub/amod.py', 'sub/bmod.py'],
    )

    assert_pequal(
        iter_files(['file.py', 'random/path'], '*/test_*,amod*'),
        [
            'file.py',
            'test_all.py',
            'tests/run.py',
            'sub/amod.py',
            'sub/bmod.py',
        ],
    )

    assert_pequal(
        iter_files(['file.py', 'random/path'], None, 'tests'),
        ['file.py', 'amod.py', 'test_all.py', 'sub/amod.py', 'sub/bmod.py'],
    )

    assert_pequal(
        iter_files(['file.py', 'random/path'], None, 'tests,sub'),
        ['file.py', 'amod.py', 'test_all.py'],
    )
    tools._is_python_file = _orig_is_python_file
    os_mod.walk = _orig_walk


CC_RESULTS_CASES = [
    (
        Function('name', 12, 0, 16, False, None, [], 6),
        {
            'type': 'function',
            'name': 'name',
            'lineno': 12,
            'col_offset': 0,
            'endline': 16,
            'closures': [],
            'complexity': 6,
            'rank': 'B',
        },
    ),
    (
        Class(
            'Classname',
            17,
            0,
            29,
            [Function('name', 19, 4, 26, True, 'Classname', [], 7)],
            [],
            7,
        ),
        {
            'type': 'class',
            'name': 'Classname',
            'lineno': 17,
            'col_offset': 0,
            'endline': 29,
            'complexity': 7,
            'rank': 'B',
            'methods': [
                {
                    'type': 'method',
                    'lineno': 19,
                    'col_offset': 4,
                    'endline': 26,
                    'closures': [],
                    'complexity': 7,
                    'rank': 'B',
                    'classname': 'Classname',
                    'name': 'name',
                }
            ],
        },
    ),
    (
        Function(
            'name',
            12,
            0,
            16,
            False,
            None,
            [Function('aux', 13, 4, 17, False, None, [], 4)],
            10,
        ),
        {
            'type': 'function',
            'name': 'name',
            'lineno': 12,
            'col_offset': 0,
            'endline': 16,
            'complexity': 10,
            'rank': 'B',
            'closures': [
                {
                    'name': 'aux',
                    'lineno': 13,
                    'col_offset': 4,
                    'endline': 17,
                    'closures': [],
                    'complexity': 4,
                    'rank': 'A',
                    'type': 'function',
                }
            ],
        },
    ),
]


@pytest.mark.parametrize('blocks,dict_result', CC_RESULTS_CASES)
def test_cc_to_dict(blocks, dict_result):
    assert tools.cc_to_dict(blocks) == dict_result


CC_TO_XML_CASE = [
    {
        'closures': [],
        'endline': 16,
        'complexity': 6,
        'lineno': 12,
        'is_method': False,
        'name': 'name',
        'col_offset': 0,
        'rank': 'B',
    },
    {
        'complexity': 8,
        'endline': 29,
        'rank': 'B',
        'lineno': 17,
        'name': 'Classname',
        'col_offset': 0,
    },
    {
        'classname': 'Classname',
        'closures': [],
        'endline': 26,
        'complexity': 7,
        'lineno': 19,
        'is_method': True,
        'name': 'name',
        'col_offset': 4,
        'rank': 'B',
    },
    {
        'closures': [],
        'endline': 17,
        'complexity': 4,
        'lineno': 13,
        'is_method': False,
        'name': 'aux',
        'col_offset': 4,
        'rank': 'A',
    },
    {
        'endline': 16,
        'complexity': 10,
        'lineno': 12,
        'is_method': False,
        'name': 'name',
        'col_offset': 0,
        'rank': 'B',
    },
]


CC_TO_CODECLIMATE_CASE = [
    {
        'closures': [],
        'endline': 16,
        'complexity': 6,
        'lineno': 12,
        'type': 'function',
        'name': 'foo',
        'col_offset': 0,
        'rank': 'B',
    },
    {
        'complexity': 8,
        'endline': 29,
        'rank': 'B',
        'lineno': 17,
        'type': 'class',
        'name': 'Classname',
        'col_offset': 0,
    },
    {
        'closures': [],
        'endline': 17,
        'complexity': 4,
        'lineno': 13,
        'type': 'method',
        'name': 'bar',
        'col_offset': 4,
        'rank': 'A',
    },
]


def test_raw_to_dict():
    assert tools.raw_to_dict(Module(103, 123, 98, 8, 19, 5, 3)) == {
        'loc': 103,
        'lloc': 123,
        'sloc': 98,
        'comments': 8,
        'multi': 19,
        'blank': 5,
        'single_comments': 3,
    }


def test_cc_to_xml():
    assert (
        tools.dict_to_xml({'filename': CC_TO_XML_CASE})
        == '''<ccm>
              <metric>
                <complexity>6</complexity>
                <unit>name</unit>
                <classification>B</classification>
                <file>filename</file>
                <startLineNumber>12</startLineNumber>
                <endLineNumber>16</endLineNumber>
              </metric>
              <metric>
                <complexity>8</complexity>
                <unit>Classname</unit>
                <classification>B</classification>
                <file>filename</file>
                <startLineNumber>17</startLineNumber>
                <endLineNumber>29</endLineNumber>
              </metric>
              <metric>
                <complexity>7</complexity>
                <unit>Classname.name</unit>
                <classification>B</classification>
                <file>filename</file>
                <startLineNumber>19</startLineNumber>
                <endLineNumber>26</endLineNumber>
              </metric>
              <metric>
                <complexity>4</complexity>
                <unit>aux</unit>
                <classification>A</classification>
                <file>filename</file>
                <startLineNumber>13</startLineNumber>
                <endLineNumber>17</endLineNumber>
              </metric>
              <metric>
                <complexity>10</complexity>
                <unit>name</unit>
                <classification>B</classification>
                <file>filename</file>
                <startLineNumber>12</startLineNumber>
                <endLineNumber>16</endLineNumber>
              </metric>
            </ccm>'''.replace(
            '\n', ''
        ).replace(
            ' ', ''
        )
    )


CC_TO_MD_RESULTS = [
    {"type": "method", "rank": "A", "lineno": 110, "classname": "Classname", "endline": 117,
     "complexity": 2, "name": "flush", "col_offset": 4},
    {"type": "class", "rank": "B", "lineno": 73, "endline": 124, "complexity": 4, "name": "Classname",
     "col_offset": 0},
    {'type': 'function', 'rank': 'C', 'lineno': 62, 'endline': 69, 'complexity': 10, 'name': 'decrement',
     'col_offset': 0}
]


def test_cc_to_md():
    md = tools.dict_to_md({'filename': CC_TO_MD_RESULTS})
    _md = '''
| Filename | Name | Type | Start:End Line | Complexity | Clasification |
| -------- | ---- | ---- | -------------- | ---------- | ------------- |
| filename | Classname.flush | M | 110:117 | 2 | A |
| filename | Classname | C | 73:124 | 4 | B |
| filename | decrement | F | 62:69 | 10 | C |
'''
    assert md == _md


def test_cc_error_to_codeclimate():
    error_result = {'error': 'Error: invalid syntax (<unknown>, line 100)'}

    expected_results = [
        json.dumps(
            {
                'description': 'Error: Error: invalid syntax (<unknown>, line 100)',
                'check_name': 'Complexity',
                'content': {
                    'body': 'We encountered an error attempting to analyze this line.'
                },
                'location': {
                    'path': 'filename',
                    'lines': {'begin': 100, 'end': 100},
                },
                'type': 'issue',
                'categories': ['Bug Risk'],
                'remediation_points': 1000000,
                'fingerprint': '10ac332cd7f638664e8865b098a1707c',
            }
        ),
    ]

    actual_results = tools.dict_to_codeclimate_issues(
        {'filename': error_result}
    )

    actual_sorted = []
    for i in actual_results:
        actual_sorted.append(json.loads(i))

    expected_sorted = []
    for i in expected_results:
        expected_sorted.append(json.loads(i))

    assert actual_sorted == expected_sorted


def test_cc_to_codeclimate():
    actual_results = tools.dict_to_codeclimate_issues(
        {'filename': CC_TO_CODECLIMATE_CASE}
    )
    expected_results = [
        json.dumps(
            {
                'description': 'Cyclomatic complexity is too high in function foo. (6)',
                'check_name': 'Complexity',
                'content': {'body': tools.get_content()},
                'location': {
                    'path': 'filename',
                    'lines': {'begin': 12, 'end': 16},
                },
                'type': 'issue',
                'categories': ['Complexity'],
                'remediation_points': 1100000,
                'fingerprint': 'afbe2b8d9a57fde5f3235ec97e7a22e1',
            }
        ),
        json.dumps(
            {
                'description': 'Cyclomatic complexity is too high in class Classname. (8)',
                'check_name': 'Complexity',
                'content': {'body': tools.get_content()},
                'location': {
                    'path': 'filename',
                    'lines': {'begin': 17, 'end': 29},
                },
                'type': 'issue',
                'categories': ['Complexity'],
                'remediation_points': 1300000,
                'fingerprint': '8caecbb525375d825b95c23bc8f881d7',
            }
        ),
    ]

    actual_sorted = []
    for i in actual_results:
        actual_sorted.append(json.loads(i))

    expected_sorted = []
    for i in expected_results:
        expected_sorted.append(json.loads(i))

    assert actual_sorted == expected_sorted


CC_TO_TERMINAL_CASES = [
    Class(
        name='Classname',
        lineno=17,
        col_offset=0,
        endline=29,
        methods=[
            Function(
                name='meth',
                lineno=19,
                col_offset=4,
                endline=26,
                is_method=True,
                classname='Classname',
                closures=[],
                complexity=4,
            )
        ],
        inner_classes=[],
        real_complexity=4,
    ),
    Function(
        name='meth',
        lineno=19,
        col_offset=4,
        endline=26,
        is_method=True,
        classname='Classname',
        closures=[],
        complexity=7,
    ),
    Function(
        name='f1',
        lineno=12,
        col_offset=0,
        endline=16,
        is_method=False,
        classname=None,
        closures=[],
        complexity=14,
    ),
    Function(
        name='f2',
        lineno=12,
        col_offset=0,
        endline=16,
        is_method=False,
        classname=None,
        closures=[],
        complexity=22,
    ),
    Function(
        name='f3',
        lineno=12,
        col_offset=0,
        endline=16,
        is_method=False,
        classname=None,
        closures=[],
        complexity=32,
    ),
    Function(
        name='f4',
        lineno=12,
        col_offset=0,
        endline=16,
        is_method=False,
        classname=None,
        closures=[],
        complexity=41,
    ),
]


def test_cc_to_terminal():
    # do the patching
    tools.LETTERS_COLORS = dict((l, '<!{0}!>'.format(l)) for l in 'FMC')
    tools.RANKS_COLORS = dict((r, '<|{0}|>'.format(r)) for r in 'ABCDEF')
    tools.BRIGHT = '@'
    tools.RESET = '__R__'

    results = CC_TO_TERMINAL_CASES
    res = [
        '@<!C!>C __R__17:0 Classname - <|A|>A (4)__R__',
        '@<!M!>M __R__19:4 Classname.meth - <|B|>B (7)__R__',
        '@<!F!>F __R__12:0 f1 - <|C|>C (14)__R__',
        '@<!F!>F __R__12:0 f2 - <|D|>D (22)__R__',
        '@<!F!>F __R__12:0 f3 - <|E|>E (32)__R__',
        '@<!F!>F __R__12:0 f4 - <|F|>F (41)__R__',
    ]
    res_noshow = ['{0}__R__'.format(r[: r.index('(') - 1]) for r in res]

    assert tools.cc_to_terminal(results, False, 'A', 'F', False) == (
        res_noshow,
        120,
        6,
    )
    assert tools.cc_to_terminal(results, True, 'A', 'F', False) == (
        res,
        120,
        6,
    )
    assert tools.cc_to_terminal(results, True, 'A', 'D', False) == (
        res[:-2],
        47,
        4,
    )
    assert tools.cc_to_terminal(results, False, 'A', 'D', False) == (
        res_noshow[:-2],
        47,
        4,
    )
    assert tools.cc_to_terminal(results, True, 'C', 'F', False) == (
        res[2:],
        109,
        4,
    )
    assert tools.cc_to_terminal(results, True, 'B', 'E', False) == (
        res[1:-1],
        75,
        4,
    )
    assert tools.cc_to_terminal(results, True, 'B', 'F', True) == (
        res[1:],
        120,
        6,
    )
